------------------------------------------------------------------------
----                                                                ----
---- SSC_8530 IP Core                                               ----
----                                                                ----
---- Description:                                                   ----
---- This model provides an asynchronous SCSI interface compa-      ----
---- tible to the DP5380 from National Semiconductor and others.    ----
----                                                                ----
---- This file covers the transceiver section of the SCC chip.      ----
----                                                                ----
----                                                                ----
---- Author(s):                                                     ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de     ----
----                                                                ----
------------------------------------------------------------------------
----                                                                ----
---- Copyright © 2015... Wolfgang Foerster - Inventronik GmbH.      ----
----                                                                ----
---- All rights reserved. No portion of this sourcecode may be      ----
---- reproduced or transmitted in any form by any means, whether    ----
---- by electronic, mechanical, photocopying, recording or          ----
---- otherwise, without my written permission.                      ----
----                                                                ----
------------------------------------------------------------------------
--
-- Revision History
--
-- Revision 2K15B 20151224 WF
--   Draft model.
-- Revision 2K20B 20201224 WF
--   Initial Release.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std-logic_arith.all;

entity TRANSCEIVER is

    port(
        CLK                 : in std_logic;

        -- Serial Data:
        RxD                 : in std_logic;
        TxD                 : out std_logic;
        TxD_EN              : out std_logic;

        -- Channel clocks:
        TRxC_INn            : in std_logic;
        TRxC_OUTn           : out std_logic;
        TRxC_EN             : out std_logic;
        RTxCn               : in std_logic;

        -- Channel controls:
        SYNC_IN             : in std_logic;
        SYNC_OUT            : out std_logic;
        SYNC_EN             : out std_logic;
        Wn_REQn             : out std_logic;

        -- Control signals:
    ONE_CLK_MISS        : out std_logic;
    TWO_CLK_MISS        : out std_logic;
LOOP_SEND           : out std_logic;
ON_LOOP             : out std_logic;

RES_CODE            : out std_logic_vector(2 downto 0);
        DCDn                : in std_logic;
        CTSn                : in std_logic;
        Rx_IP               : buffer std_logic;
        Tx_IP               : buffer std_logic;
        EXT_STAT_IP         : buffer std_logic;

        BREAK_ABORT_IE      : in std_logic:
        EOM_IE              : in std_logic:
        CTS_IE              : in std_logic:
        SYNC_HUNT_IE        : in std_logic:
        DCD_IE              : in std_logic:
        FRAME_FIFO_EN       : in std_logic:
        ZCOUNT_IE           : in std_logic;
        SDLC_HDLC           : in std_logic;
        DPLL_COMMAND        : in std_logic_vector(2 downto 0);
        LOOPBACK            : in std_logic;
        AUTO_ECHO           : in std_logic;
        BR_GEN_SRC          : in std_logic;
        BR_GEN_EN           : in std_logic;
        XTAL                : in std_logic;
        Rx_CLK_SEL          : in std_logic_vector(1 downto 0);
        Tx_CLK_SEL          : in std_logic_vector(1 downto 0);
        TRXCn_SEL           : in std_logic_vector(2 downto 0);
        NRZ_FM              : in std_logic_vector(1 downto 0);
ROLL                : in std_logic;
MARK_FLAGn          : in std_logic;
ABORT_FLAGn         : in std_logic;
LOOP_MODE           : in std_logic;
B6_B8n              : in std_logic;
        RES                 : in std_logic;
AUTO_Tx_FLAG        : in std_logic;
AUTO_EOM_LRES       : in std_logic;
        AUTO_RTS            : in std_logic;
        TxD_PULLED_HIGH     : in std_logic;
        FAST_DTR            : in std_logic;
CRC_CHECK_RCVD      : in std_logic;
EXTEND_READ_EN      : in std_logic;
        SYNC_SDLC_ADR       : in std_logic_vector(7 downto 0);
        SYNC_SDLC_FLAG      : in std_logic_vector(7 downto 0);
        Tx_CRC_EN           : in std_logic;
        RTS                 : in std_logic;
        RTSn                : out std_logic;
CRC16_SDLCn         : in std_logic;
        Tx_EN               : in std_logic;
        SEND_BREAK          : in std_logic;
        Tx_BITS             : in std_logic_vector(1 downto 0);
        DTRn_REQn           : out std_logic;
        PAR_EN              : in std_logic;
        PAR_EVEN_ODDn       : in std_logic;
        SYNC_MODE           : in std_logic_vector(1 downto 0);
        SYNC_CHAR           : in std_logic_vector(1 downto 0);
        CLK_MODE            : in std_logic_vector(1 downto 0);
        EXT_INT_EN          : in std_logic;
        Tx_INT_EN           : in std_logic;
        PAR_S_COND          : in std_logic;
        Rx_INT_MODE         : in std_logic_vector(1 downto 0);
        DTRn_REQ            : in std_logic;
        DMA_REQ_MODE        : in std_logic_vector(2 downto 0);
        Rx_EN               : in std_logic;
SYNC_CHAR_INH       : in std_logic;
ADR_SEARCH_MODE     : in std_logic;
Rx_CRC_EN           : in std_logic;
        ENTER_HUNT_MODE     : in std_logic;
        AUTO_EN             : in std_logic;
        Rx_BITS             : in std_logic_vector(1 downto 0);
        TC                  : in std_logic_vector(15 downto 0);
        INTACKn_INH         : in std_logic;
        MIE                 : in std_logic;
        DLC                 : in std_logic;
        INT_ACK             : in std_logic;
        IEI                 : in std_logic;
        INT                 : out std_logic;
        STATUS              : out std_logic_vector(1 downto 0);
        RESET               : in std_logic;

        SEND_ABORT          : in std_logic;
        EN_INT_RxCHAR       : in std_logic;
        RES_EXT_STAT_INT    : in std_logic;
        RES_TxINT           : in std_logic;
        RES_ERR             : in std_logic;
        RES_IUS             : in std_logic;
        CRC_PRES            : in std_logic;
        RES_Rx_CRC          : in std_logic;
        RES_Tx_CRC          : in std_logic;
        RES_Tx_UR_EOM       : in std_logic;

        BUFFER_IN           : in std_logic_vector(7 downto 0);
        BUFFER_OUT          : out std_logic_vector(7 downto 0);
        RR0_RD              : in std_logic;
        RR1_RD              : in std_logic;
        RR6_RD              : in std_logic;
        RR7_RD              : in std_logic;
        RR8_RD              : in std_logic;
        WR8_WR              : in std_logic
    );
end entity TRANSCEIVER;

architecture BEHAVIOUR of TRENSCEIVER is
type CLK_MULTIPLIER is(64, 32, 16, 1);
type OPMODES is(ASYNC, MSYNC, BSYNC, SDLC);
type FIFOTYPE is array(0 to 3) of std_logic_vector(15 downto 0);
type FRAME_FIFO_TYPE is array(0 to 9) of std_logic_vector(18 downto 0);
type Rx_STATES is(IDLE, Rx_HUNT, Rx_CLKSYNC, Rx_SHIFTIN, Rx_CHECK_FRAME):
type Tx_STATES is(IDLE, Tx_START, Tx_SHIFTOUT, Tx_STOP_1, Tx_STOP_2):
signal Rx_STATE             : Rx_STATES;
signal Rx_NSTATE            : Rx_STATES; -- Next state.
signal Tx_STATE             : Tx_STATES;
signal Tx_NSTATE            : Tx_STATES; -- Next state.
signal FIFO_REG             : FIFOTYPE;
signal FIFO_WR_PNT          : natural range 0 to 3;
signal FRAME_FIFO           : FRAME_FIFO_TYPE;
signal FRAME_FIFO_DIN       : std_logic_vector(18 downto 0);
signal FRAME_FIFO_DOUT      : std_logic_vector(18 downto 0);
signal FRAME_FIFO_ADR       : std_logic_vector(3 downto 0);
signal FRAME_FIFO_HEAD      : std_logic_vector(3 downto 0);
signal FRAME_FIFO_TAIL      : std_logic_vector(3 downto 0);
signal FRAME_FIFO_WR        : std_logic;
signal FRAME_FIFO_RD        : std_logic;
signal FRAME_FIFO_EMPTY     : std_logic;
signal FRAME_FIFO_OVERFLOW  : std_logic;
signal CLK_MUL              : CLK_MULTIPLIER;
signal OPMODE               : OPMODES;
signal ABORT                : std_logic;
signal BREAK                : std_logic;
signal BRG_OUT              : std_logic;
signal EXT_STATUS_INT       : std_logic;
signal IUS                  : std_logic;
signal Rx_DPLL_OUT          : std_logic;
signal Tx_DPLL_OUT          : std_logic;
signal Rx_CHAR              : std_logic;
signal Rx_CHAR_1            : std_logic;
signal Rx_CLK               : std_logic;
signal Rx_INH               : std_logic;
signal RxD_D                : std_logic;
signal RxD_I                : std_logic;
signal RxD_NRZI             : std_logic;
signal RxD_SR_3B            : std_logic;
signal RxD_SR               : std_logic;
signal RxD_SDLC             : std_logic;
signal Rx_SCOND             : std_logic;
signal SDLC_MODE            : boolean;
signal Tx_CLK               : std_logic;
signal Tx_INH               : std_logic;
signal Tx_SR_DATAIN         : std_logic_vector(19 downto 0);
signal TR_EMPTY             : std_logic;
signal TR_FLAG_SENT         : std_logic;
signal TxD_I                : std_logic;
signal RTxC_I               : std_logic;
signal TRxC_I               : std_logic;
signal Tx_BUFFER_EMPTY      : std_logic;
signal Rx_BUFFER_EMPTY      : std_logic;
signal EOF                  : std_logic;
signal CRC_FRAME_ERR        : std_logic;
signal Rx_OVR               : std_logic;
signal FIFO_PAR_ERR         : std_logic;
signal PARITY_ERR           : std_logic;
signal RESIDUE_BITS         : std_logic_vector(2 downto 0);
signal FRAME_BYTE_COUNTER   : std_logic_vector(13 downto 0);
signal Rx_SHIFTREG          : std_logic_vector(8 downto 0); -- Data and parity.
signal Rx_SHFT_RDY          : std_logic;
signal Tx_SHFT_RDY          : std_logic;
signal TxD_SR               : std_logic;
signal Tx_CRC               : std_logic;
signal Tx_ZI                : std_logic;
signal Rx_ERROR             : std_logic_vector(7 downto 0);
signal Rx_PARITY            : std_logic;
signal Rx_PARBIT            : std_logic;
signal FIFO_RD              : std_logic;
signal FIFO_WR              : std_logic;
signal FIFO_FULL            : std_logic;
signal FIFO_EMPTY           : std_logic;
signal Rx_CHAR_AV           : std_logic;
signal ZCOUNT               : std_logic;
signal Tx_MUX               : std_logic;
signal DCD                  : std_logic;
signal SYNC_HUNT            : std_logic;
signal CTS                  : std_logic;
signal Tx_UNDERRUN_EOM      : std_logic;
signal BREAK_ABORT          : std_logic;
signal Tx_LEN               : integer range 1 to 8;
signal SYNC_SDLC_ADR_I      : std_logic_vector(7 downto 0);
signal SYNC_SDLC_FLAG_I     : std_logic_vector(7 downto 0);
signal ALL_SENT             : std_logic;
signal Rx_STRB              : std_logic;
signal Tx_STRB              : std_logic;
begin
    P_SYNC: process
    begin
        wait until CLK = '1' and CLK' event;
        case AUTO_EN = '1' and LOOPBACK = '0' is
            when '1' =>
                Rx_INH <= DCDn;
                Tx_INH <= CTSn;
            when others => -- Always on.
                Rx_INH <= '0';
                Tx_INH <= '0';
        end case;

        if AUTO_ECHO = '1' then
            Tx_INH <= '0';
        end if;

        RxD_D <= RxD; -- One bit input synchronization.
        RTxC_I <= not RTxCn;
        TRxC_I <= not TRxC_INn;
    end process P_SYNC;

    SYNC_OUT <= not RTxCn when XTAL = '1' else -- This is the crystal oscillator feedback.
                '1' when Rx_STATE = Rx_HUNT else '0';

    SYNC_EN <= '1' when XTAL = '1' else -- Crystal oscillator is enabled.
               '1' when OPMODE /= ASYNC and SYNC_CHAR /= "11"; -- Synchronous mode except external sync.

    -- Echo mode:
    TRxC_EN <= TRXCn_SEL(2);

    with TRXCn_SEL(1 downto 0) select
        TRxC_OUTn <= not RTxC_I when "00",
                     not Tx_CLK when "01",
                     not BRG_OUT when "10",
                     not Rx_DPLL_OUT when others;

    CLK_MUL <= 1 when OPMODE /= ASYNC and SYNC_CHAR /= "11" else -- Synchronous mode except external sync.
               64 when CLK_MODE = "11" else
               32 when CLK_MODE = "10" else
               16 when CLK_MODE = "01" else 1;

    with Rx_CLK_SEL select
        Rx_CLK <= RTxC_I when "00",
                  TRxC_I when "01",
                  BRG_OUT when "10",
                  Rx_DPLL_OUT when others;

    with Tx_CLK_SEL select
        Tx_CLK <= RTxC_I when "00",
                  TRxC_I when "01",
                  BRG_OUT when "10",
                  Tx_DPLL_OUT when others;

    Rx_STROBE: process
    -- This is the clock strobe for the receive shift register.
    -- The data is sampled on the rising edge of the receive
    -- clock. The sampling time is in the middle of the bit cell
    -- when the clock multiplier is 16, 32 or 64. In case of a
    -- clock multiplier of 1 the strobe reflects the rising edge
    -- of the receiver clock.
    variable LOCK           : boolean;
    variable Rx_PRESCALE    : std_logic_vector(5 downto 0);
    begin
        wait until CLK = '1' and CLK' event;
        Rx_STRB <= '0';
        if Rx_STATE = IDLE then
            Rx_PRESCALE := "00000";
        else
            if Rx_CLK = '1' and LOCK = false then
                LOCK := true;
                case CLK_MUL is
                    when 64 =>
                        Rx_PRESCALE := Rx_PRESCALE + '1';
                        if Rx_PRESCALE = "011111" then
                            Rx_STRB <= '1';
                        end if;
                    when 32 =>
                        if Rx_PRESCALE = "001111" then
                            Rx_STRB <= '1';
                            Rx_PRESCALE := Rx_PRESCALE + '1';
                        elsif Rx_PRESCALE = "011111" then
                            Rx_PRESCALE <= "000000";
                        end if;
                    when 16 =>
                        if Rx_PRESCALE = "000111" then
                            Rx_STRB <= '1';
                            Rx_PRESCALE := Rx_PRESCALE + '1';
                        elsif Rx_PRESCALE = "001111" then
                            Rx_PRESCALE <= "000000";
                        end if;
                    when 1 =>
                        Rx_STRB <= '1';
                end case;
            elsif Rx_CLK = '0' then
                LOCK := false;
            end if;
    end process Rx_STROBE;

    Tx_STROBE: process
    -- This is the clock strobe for the transmit shift register.
    -- The data is shifted on the falling edge of the transmit
    -- clock.
    variable LOCK           : boolean;
    variable Tx_PRESCALE    : std_logic_vector(5 downto 0);
    begin
        wait until CLK = '1' and CLK' event;
        Tx_STRB <= '0';
        if Tx_STATE = IDLE then
            Tx_PRESCALE := "00000";
        else
            if Tx_CLK = '0' and LOCK = false then
                LOCK := true;
                case CLK_MUL is
                    when 64 =>
                        Tx_PRESCALE := Tx_PRESCALE + '1';
                        if Tx_STATE = Tx_STOP_2 and SYNC_MODE_A = "10" and Tx_PRESCALE = "011111" then -- 1.5 stop bits.
                            Tx_STRB <= '1';
                        elsif Tx_PRESCALE = "111111" then
                            Tx_STRB <= '1';
                        end if;
                    when 32 =>
                        if Tx_STATE = Tx_STOP_2 and SYNC_MODE_A = "10" and Tx_PRESCALE = "001111" then -- 1.5 stop bits.
                            Tx_STRB <= '1';
                        elsif Tx_PRESCALE = "011111" then
                            Tx_STRB <= '1';
                            Tx_PRESCALE <= "000000";
                        else
                            Tx_PRESCALE := Tx_PRESCALE + '1';
                        end if;
                    when 16 =>
                        if Tx_STATE = Tx_STOP_2 and SYNC_MODE_A = "10" and Tx_PRESCALE = "000111" then -- 1.5 stop bits.
                            Tx_STRB <= '1';
                        elsif Tx_PRESCALE = "001111" then
                            Tx_STRB <= '1';
                            Tx_PRESCALE <= "000000";
                        else
                            Tx_PRESCALE := Tx_PRESCALE + '1';
                        end if;
                    when 1 =>
                        Tx_STRB <= '1';
                end case;
            elsif Tx_CLK = '1' then
                LOCK := false;
            end if;
    end process Tx_STROBE;

    BAUDRATE_GENERATOR: process
    -- This is the SCCs baude rate generator.
    variable BRG_CNT    : std_logic_vector(15 downto 0);
    variable LOCK       : boolean;
    variable STARTLOCK  : boolean;
    begin
        wait until CLK = '1' and CLK' event;
        if (RESET or RES) = '1' then
            ZCOUNT <= '0';
        elsif BR_GEN_EN = '0' then
            BRG_CNT := TC;
            ZCOUNT <= '0';
            BRG_OUT <= '1';
            LOCK := true;
            STARTLOCK := false;
        elsif BR_GEN_SRC = '0' and RTxC_I = '0' and STARTLOCK = false then
                STARTLOCK := true; -- Wait for initial condition.
        elsif BR_GEN_SRC = '0' and STARTLOCK = true then -- Use RTxC as clock source.
            if RTxC_I = '1' and LOCK = false and BRG_CNT > x"0000" then
                BRG_CNT := BRG_CNT - '1';
                ZCOUNT <= '0';
                LOCK := true;
            elsif RTxC_I = '1' and LOCK = false then
                BRG_CNT := TC;
                ZCOUNT <= '1';
                BRG_OUT <= not BRG_OUT;
                LOCK := true;
            elsif RTxC_I = '0' then
                LOCK := false;
            end if;
        elsif BR_GEN_SRC = '1' -- Use CLK as clock source.
            if BRG_CNT > x"0000" then
                BRG_CNT := BRG_CNT - '1';
                ZCOUNT <= '0';
            else
                BRG_CNT := TC;
                ZCOUNT <= '1';
                BRG_OUT <= not BRG_OUT;
            end if;
        end if;

        if ZCOUNT_IE = '0' then
            ZCOUNT <= '0'; -- Disable ZCOUNT interrupt source.
        end if;
    end process BAUDRATE_GENERATOR;

    P_DPLL process
    -- This is the digital phase locked loop section of the SCC.
    -- It is provided to handle the baud rate clock or the receiver
    -- clock smaller than the system clock. For 32MHz receiver
    -- clock operation stated out in the original data sheet, the core
    -- clock is for example 48MHz or 64 MHz. The reason for this is
    -- to meet the rules for setup / hold times which are handled in
    -- this case in one common clock domain and to avoid asynchronous
    -- settings and latches.

    -- The DPLL commands are as follows:
    -- "000" = Null command.
    -- "001" = Enter search mode.
    -- "010" = Reset missing clock.
    -- "011" = Disable DPLL.
    -- "100" = Set source = baud rate generator.
    -- "101" = Set source = RTxC.
    -- "110" = Set FM mode.
    -- "111" = Set NRZI mode.
    variable SRC_BRG_RTxCn  : std_logic;
    variable MODE_FM_NRZIn  : std_logic;
    variable DPLL_EN        : std_logic;
    variable EDGE_DETECT    : std_logic_vector(1 downto 0);
    variable SEARCHMODE     : std_logic;
    variable LOCK           : std_logic;
    variable SWITCH         : std_logic;
    variable COUNTER        : std_logic_vector(4 downto 0);
    variable MODIFIER       : std_logic_vector(1 downto 0);
    variable FM_EDGE        : std_logic;
    begin
        wait until CLK = '1' and CLK' event;
        if RESET = '1' or RES = '1' then
            DPLL_EN := '0';
        end if;

        case DPLL_COMMAND is
            when "000" => -- Null command.
                if SWITCH = '1' then
                    DPLL_EN := '1'; -- Switch right in the end of command "001".
                end if;
                SWITCH := '0';
            when "001" => -- Enter search mode.
                SWITCH := '1';
                if DPLL_EN = '1' then
                    SEARCHMODE := '1';
                end if;
                ONE_CLK_MISS <= '0';
                TWO_CLK_MISS <= '0';
            when "010" => -- Reset missing clock.
                ONE_CLK_MISS <= '0';
                TWO_CLK_MISS <= '0';
            when "011" => -- Disable DPLL.
                DPLL_EN := '0';
            when "100" => -- Set source = baud rate generator.
                SRC_BRG_RTxCn := '1';
            when "101" => -- Set source = RTxC.
                SRC_BRG_RTxCn := '0';
            when "110" => -- Set FM mode.
                MODE_FM_NRZIn := '1';
            when "111" => -- Set NRZI mode.
                MODE_FM_NRZIn := '1';
            when others =>
                null; -- Do nothing.
        end case;

        if DPLL_EN = '0' then
            EDGE_DETECT := "00";
            SEARCHMODE := '0';
            Rx_DPLL_OUT <= '0';
            Tx_DPLL_OUT <= '0';
            COUNTER := "00000";
            MODIFIER := "00";
            ONE_CLK_MISS <= '0';
            TWO_CLK_MISS <= '0';
            SRC_BRG_RTxCn := '0';
            MODE_FM_NRZIn := '0';
        elsif SEARCHMODE = '1' and (EDGE_DETECT(1) xor EDGE_DETECT(0)) = '1' then -- Detect positive or negative edge.
            SEARCHMODE := '0';
            Rx_DPLL_OUT <= '1';
            Tx_DPLL_OUT <= '1';
            COUNTER := "10000";
            MODIFIER := "00";
        elsif SEARCHMODE = '1' then
            EDGE_DETECT := EDGE_DETECT(0) & RxD_D;
            COUNTER := "10000";
            MODIFIER := "00";
            FM_EDGE := '0';
            if MODE_FM_NRZIn = '0' then -- NRZI.
                Rx_DPLL_OUT <= '1';
                Tx_DPLL_OUT <= '1';
            else -- FM.
                Rx_DPLL_OUT <= '0';
                Tx_DPLL_OUT <= '0';
            end if;
        else
            if (EDGE_DETECT(1) xor EDGE_DETECT(0)) = '1' then -- Adjust phase.
                if MODE_FM_NRZIn = '0' then -- NRZI.
                    if COUNTER > "10000" then
                        MODIFIER := "01";  -- Add one.
                    elsif COUNTER < "01111" then
                        MODIFIER := "11";  -- Subtract one.
                    end if;
                else -- FM.
                    case COUNTER is
                        when "10001" | "10010" | "10011" =>
                            MODIFIER := "01";  -- Add one.
                        when "01100" | "01101" | "01110" =>
                            MODIFIER := "11";  -- Subtract one.
                        when others => null;
                    end case;
                end if;
            elsif (SRC_BRG_RTxCn = '1' and BRG_OUT = '1') or (SRC_BRG_RTxCn = '0' and RTXC_I = '1') and LOCK = false then
                if MODIFIER = "01" and COUNTER = "00100" then -- Add.
                    COUNTER := "00110"; -- Delete count five.
                    MODIFIER := "00";
                elsif MODIFIER = "11" and COUNTER = "00101" then -- Sub.
                    COUNTER := "00101"; -- Double count five.
                    MODIFIER := "00";
                else
                    COUNTER := COUNTER + '1';
                end if;

                if MODE_FM_NRZIn = '0' then -- This logic is exclusively for FM.
                    SEARCHMODE := '0';
                elsif COUNTER >= "01100" and COUNTER < "10011" and (EDGE_DETECT(1) xor EDGE_DETECT(0)) = '1' then
                    FM_EDGE := '1'; -- Edge detected.
                elsif COUNTER = "10100" and FM_EDGE = '0' and ONE_CLK_MISS = '0' then -- No edge detected.
                    ONE_CLK_MISS <= '1'; -- First missing edge.
                    FM_EDGE := '0'; -- Re-initialize.
                elsif COUNTER = "11111" and FM_EDGE = '0' then -- No edge detected.
                    SEARCHMODE := '1'; -- Two bitcells without edge, enter searchmode.
                    TWO_CLK_MISS <= '1';
                end if;
                LOCK := true;
            elsif (SRC_BRG_RTxCn = '1' and BRG_OUT = '0') or (SRC_BRG_RTxCn = '0' and RTXC_I = '0') then
                LOCK := false;
            end if;

            if MODE_FM_NRZIn = '0' then -- NRZI.
                if COUNTER < "10000" then
                    Rx_DPLL_OUT <= '0';
                    Tx_DPLL_OUT <= '0';
                else
                    Rx_DPLL_OUT <= '1';
                    Tx_DPLL_OUT <= '1';
                end if;
            else -- FM.
                case COUNTER is
                    when "00000" | "00001" | "00010" | "00011" | "10000" | "10001" | "10010" | "10011" =>
                        Rx_DPLL_OUT <= '0';
                        Tx_DPLL_OUT <= '0';
                    when "00100" | "00101" | "00110" | "00111" | "10100" | "10101" | "10110" | "10111" =>
                        Rx_DPLL_OUT <= '1';
                        Tx_DPLL_OUT <= '0';
                    when "01000" | "01001" | "01010" | "01011" | "11000" | "11001" | "11010" | "11011" =>
                        Rx_DPLL_OUT <= '1';
                        Tx_DPLL_OUT <= '1';
                    when others =>
                        Rx_DPLL_OUT <= '0';
                        Tx_DPLL_OUT <= '1';
                end case;
            end if;

            EDGE_DETECT := EDGE_DETECT(0) & RxD_D;
        end if;
    end process P_DPLL;

    STATE_REGISTERS: process
    begin
        wait until CLK = '1' and CLK' event;
        Rx_STATE <= Rx_NSTATE;
        Tx_STATE <= Tx_NSTATE;
    end process STATE_REGISTERS;

Rx_STATE_DECODER: process(Rx_STATE, Rx_EN, OPMODE, RxD, RxD_D, Rx_STRB, Rx_SHFT_RDY)
begin
    case Rx_STATE is
        when IDLE =>
            if Rx_EN = '0' then
                Rx_NSTATE <= IDLE;
            elsif OPMODE = ASYNC and RxD = '0' RxD_D = '1' then -- Falling edge detected.
                Rx_NSTATE <= Rx_CLKSYNC;
            elsif OPMODE = SDLC and ENTER_HUNT_MODE = '1' then
                Rx_NSTATE <= Rx_HUNT;
            else
                Rx_NSTATE <= IDLE;
            end if;
when Rx_HUNT =>
    if 
        when Rx_CLKSYNC => -- Wait half a bit cell.
            if OPMODE = ASYNC and Rx_STRB = '1' and RxD_D = '1' then
                Rx_NSTATE <= IDLE; -- Not a start bit.
            elsif OPMODE = ASYNC and Rx_STRB = '1' then
                Rx_NSTATE <= Rx_SHIFTIN;
            else
                Rx_NSTATE <= Rx_CLKSYNC;
            end if;
        when Rx_SHIFTIN =>
            if OPMODE = ASYNC and Rx_SHFT_RDY then
                Rx_NSTATE <= Rx_CHECK_FRAME;
            else
                Rx_NSTATE <= Rx_SHIFTIN;
            end if;
        when Rx_CHECK_FRAME =>
            if OPMODE = ASYNC and Rx_STRB = '1' then
                Rx_NSTATE <= IDLE;
            else
                Rx_NSTATE <= Rx_CHECK_FRAME;
            end if;
    end case;
end process Rx_STATE_DECODER;

Tx_STATE_DECODER: process(Tx_STATE)
    begin
        case Tx_STATE is
            when IDLE =>
                if Tx_EN = '0' then
                    Tx_NSTATE <= IDLE;
                elsif OPMODE = ASYNC and Tx_BUFFER_EMPTY = '0' then
                    Tx_NSTATE <= Tx_START;
                else
                    Tx_NSTATE <= IDLE;
                end if;
            when Tx_START =>
                if Tx_STRB = '1' then
                    Tx_NSTATE <= Tx_SHIFTOUT;
                else
                    Tx_NSTATE <= Tx_START;
                end if;
            when Tx_SHIFTOUT =>
                if OPMODE = ASYNC and Rx_SHFT_RDY = '1' then
                    Tx_NSTATE <= Tx_STOP_1;
                else
                    Tx_NSTATE <= Tx_STOP_1;
                end if;
            when Tx_STOP_1 =>
                elsif Tx_STRB = '1' and SYNC_MODE_A = "01" then -- One stop bit.
                    Tx_NSTATE <= IDLE;
                if Tx_STRB = '1' then -- 1.5 or 2 stop bits.
                    Tx_NSTATE <= Tx_STOP_2;
                else
                    Tx_NSTATE <= Tx_STOP_1;
                end if;
            when Tx_STOP_2 =>
                if Tx_STRB = '1' then
                    Tx_NSTATE <= IDLE;
                else
                    Tx_NSTATE <= Tx_STOP_2;
                end if;
        end case;
    end process Tx_STATE_DECODER;

    with SYNC_MODE & SYNC_CHAR select
        OPMODE <= SDLC when "1000", -- SDLC/HDLC mode.
                  MSYNC when "0000", -- MONOSYNC mode.
                  BSYNC when "0100", -- BISYNC mode.
                  ASYNC when others; -- Asynchronous mode.

    P_RTS: process
    begin
        wait until CLK = '1' and CLK' event;
        if RESET = '1' then
            RTSn <= '1';
        elsif RTS = '1' then
            RTSn <= '0';
        elsif OPMODE = ASYNC and AUTO_EN = '1' and TR_EMPTY = '1' then -- Auto enable in asynchronous mode.
            RTSn <= '1';
        elsif OPMODE = ASYNC and AUTO_EN = '0' and RTS = '0' then -- SYNC mode or asynchronous mode without auto enable.
            RTSn <= '1';
        elsif OPMODE = SDLC and SDLC_HDLC = '1' and AUTO_RTS = '1' and TR_FLAG_SENT then -- SDLC mode.
            RTSn <= '1';
        elsif OPMODE = SDLC and SDLC_HDLC = '1' and AUTO_RTS = '0' then -- SDLC mode.
            null; -- Wait.
        elsif OPMODE /= ASYNC and RTS = 0 then
            RTSn <= '1';
        end if;
    end process P_RTS;

    P_INTERRUPTS: process
    variable IP_Rx_D        : std_logic;
    variable IP_Tx_D        : std_logic;
    variable IP_EXT_STAT_D  : std_logic;
    begin
        wait until CLK = '1' and CLK' event;

        if (RESET or RES) = '1' or MIE = '0' then
            Rx_IP <= '0';
        elsif Rx_INT_MODE = "11" and Rx_SCOND = '1' then
            Rx_IP <= '1';
        elsif Rx_INT_MODE = "10" and (Rx_CHAR = '1' or Rx_SCOND = '1') then
            Rx_IP <= '1';
        elsif Rx_INT_MODE = "01" and (Rx_CHAR_1 = '1' or Rx_SCOND = '1') then
            Rx_IP <= '1';
        elsif Rx_SCOND = '0' and Rx_CHAR = '0' and Rx_CHAR_1 = '0' then
            Rx_IP <= '0';
        end if;

        if (RESET or RES) = '1' or MIE = '0' then
            Tx_IP <= '0';
        elsif Tx_INT_EN = '1' then
            Tx_IP <= Tx_BUFFER_EMPTY;
        end if;

        if (RESET or RES) = '1' or MIE = '0' or EXT_INT_EN = '0' then
            EXT_STAT_IP <= '0';
        else
            EXT_STAT_IP <= EXT_STATUS_INT;
        end if;

        if (RESET or RES) = '1' or MIE = '0' then
            IUS <= '0'; -- Interrupt under service.
        elsif IEI = '0' and DLC = '0' then
            IUS <= '0'; -- No service at all.
        elsif INT_ACK = '1' then
            IUS <= '1'; -- IUS with interrupt acknowledge cycle.
        elsif (Rx_IP or Tx_IP or EXT_STAT_IP) = '1' and INTACKn_INH = '1' then
            IUS <= '1'; -- IUS without interrupt acknowledge cycle.
            IP_Rx_D := IP_Rx;
            IP_Tx_D := IP_Tx;
            IP_EXT_STAT_D := IP_EXT_STAT;
        elsif RES_IUS = '1' then
            IUS <= '0'; -- Clear.
        elsif IP_Rx_D = '1' and IP_Rx = '0' and INTACKn_INH = '1' then
            IUS <= '0'; -- Clear.
            IP_Rx_D := '0';
        elsif IP_Tx_D = '1' and IP_Tx = '0' and INTACKn_INH = '1' then
            IUS <= '0'; -- Clear.
            IP_Tx_D := '0';
        elsif IP_EXT_STAT_D = '1' and IP_EXT_STAT = '0' and INTACKn_INH = '1' then
            IUS <= '0'; -- Clear.
            IP_EXT_STAT_D := '0';
        end if

        if (RESET or RES) = '1' or MIE = '0' then
            INT <= '0';
        elsif INT_ACK = '1' then
            INT <= '0';
        elsif IEI = '0' and DLC = '0' then
            INT <= '0'; -- Daisy chain: inhibit interrupt processing.
        elsif IUS = '1' then
            null; -- Store the current status.
        elsif Rx_INT_MODE = "11" and Rx_SCOND = '1' then
            INT <= '1'; -- Error on special condition.
        elsif Rx_INT_MODE = "10" and (Rx_CHAR = '1' or Rx_SCOND = '1') then
            INT <= '1'; -- Interrupt on all Rx characters or special condition.
        elsif Rx_INT_MODE = "01" and (Rx_CHAR_1 = '1' or Rx_SCOND = '1') then
            INT <= '1'; -- Interrupt on first character or special condition.
        elsif Tx_INT_EN = '1' and Tx_BUFFER_EMPTY = '1' then
            INT <= '1'; -- Transmit buffer empty.
        elsif EXT_INT_EN = '1' and EXT_STATUS_INT = '1' then
            INT <= '1'; -- External or status interrupt.
        end if;

        if IUS = '1' then
            null; -- Store the current status.
        elsif Rx_SCOND = '1' then
            STATUS <= "11";
        elsif Rx_CHAR = '1' or Rx_CHAR_1 = '1' then
            STATUS <= "10";
        elsif Tx_BUFFER_EMPTY = '1' then
            STATUS <= "00";
        elsif EXT_STATUS_INT = '1' then
            STATUS <= "01";
        end if;
    end process P_INTERRUPTS;

    EXTERNAL_INTERRUPT_LATCHES: process
    variable CMD_LOCK           : boolean;
    variable LATCH_LOCK         : boolean;

    variable BREAK_ABORT_D      : std_logic;
    variable CTS_D              : std_logic;
    variable DCD_D              : std_logic;
    variable SYNC_HUNT_D        : std_logic;
    variable Tx_UNDERRUN_EOM_D  : std_logic;
    variable ZCOUNT_D           : std_logic;

    variable BREAK_ABORT_HEAD   : std_logic;
    variable BREAK_ABORT_TAIL   : std_logic;
    variable BREAK_ABORT_LATCH  : std_logic;
variable DCD_LATCH          : std_logic;
variable EOM_LATCH          : std_logic;

    variable CTS_LH             : boolean;
    variable CTS_HL             : boolean;
    variable SYNC_HUNT_LH       : boolean;
    variable SYNC_HUNT_HL       : boolean;
    begin
        wait until CLK = '1' and CLK' event;
CMD_LOCK
        if EXT_INT_EN = '0' then
            BREAK_ABORT_HEAD := false;
            BREAK_ABORT_TAIL := false;
            BREAK_ABORT_LATCH <= '0';
            DCD_LATCH <= '0';
            EOM_LATCH <= '0';
            LATCH_LOCK := false;
        elsif RES_EXT_STAT_INT = '1' then
            DCD_LATCH <= '0';
            EOM_LATCH <= '0';
            if BREAK_ABORT_HEAD = true then
                BREAK_ABORT_LATCH <= '1';
                EXT_STATUS_INT <= '1';
                BREAK_ABORT_HEAD := false;
            elsif BREAK_ABORT_TAIL = true then
                BREAK_ABORT_LATCH <= '0';
                EXT_STATUS_INT <= '1';
                BREAK_ABORT_TAIL := false;
            else
                BREAK_ABORT_LATCH <= '0';
                EXT_STATUS_INT <= '0';
            end if;
            LATCH_LOCK := false;
        elsif LATCH_LOCK = true then -- The latches are closed.
            if BREAK_ABORT_IE = '1' and OPMODE = ASYNC and BREAK_ABORT_LATCH = '1' and BREAK = '0' then
                BREAK_ABORT_TAIL := true;
            elsif BREAK_ABORT_IE = '1' and OPMODE = ASYNC and BREAK_ABORT_LATCH = '0' and BREAK = '1' then
                BREAK_ABORT_HEAD := true;
            elsif BREAK_ABORT_IE = '1' and OPMODE = SDLC and BREAK_ABORT_LATCH = '1' and ABORT = '0' then
                BREAK_ABORT_TAIL := true;
            elsif BREAK_ABORT_IE = '1' and OPMODE = SDLC and BREAK_ABORT_LATCH = '0' and ABORT = '1' then
                BREAK_ABORT_HEAD := true;
            end if;

            if OPMODE = ASYNC then -- Asynchronous sync mode.
                BREAK_ABORT_LATCH <= BREAK;
            elsif OPMODE = SDLC then
                BREAK_ABORT_LATCH <= ABORT;
            end if;
        elsif BREAK_ABORT_IE = '1' and OPMODE = ASYNC and BREAK_ABORT_LATCH = '1' and BREAK = '0' then
            BREAK_ABORT_LATCH <= '0';
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif BREAK_ABORT_IE = '1' and OPMODE = ASYNC and BREAK_ABORT_LATCH = '0' and BREAK = '1' then
            BREAK_ABORT_LATCH <= '1';
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif BREAK_ABORT_IE = '1' and OPMODE = SDLC and BREAK_ABORT_LATCH = '1' and ABORT = '0' then
            BREAK_ABORT_LATCH <= '0';
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif BREAK_ABORT_IE = '1' and OPMODE = SDLC and BREAK_ABORT_LATCH = '0' and ABORT = '1' then
            BREAK_ABORT_LATCH <= '0';
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif SYNC_HUNT_IE = '1' and SYNC_HUNT = '1' and SYNC_HUNT_D = '0' and SYNC_HUNT_HL = '0' then -- Rising edge detected.
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif SYNC_HUNT_IE = '1' and SYNC_HUNT = '0' and SYNC_HUNT_D = '1' and SYNC_HUNT_LH = '0' then -- Falling edge detected.
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif ZCOUNT_IE = '1' and ZCOUNT_D = '0' and ZCOUNT = '1' then -- Rising edge.
            -- ZCOUNT_LATCH: ZCOUNT is not latched.
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif EOM_IE = '1' and (OPMODE = MSYNC or OPMODE = BSYNC or OPMODE = SDLC) and Tx_BUFFER_EMPTY = '1' then
            EOM_LATCH <= '1';
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif CTS_IE = '1' and CTS xor CTS_LATCH = '1' then -- Any edge detected.
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif CTS_IE = '1' and CTS = '0' and CTS_D = '1' and CTS_LH = '0' then -- Falling edge detected.
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        elsif DCD_IE = '1' and DCD_LATCH = CTSn then -- Edge detected.
            DCD_LATCH <= not DCDn
            EXT_STATUS_INT <= '1';
            LATCH_LOCK := true;
        else -- LATCH_LOCK = false.
CTS_LATCH <= CTS;


if OPMODE = ASYNC then
    BREAK_ABORT_LATCH <= BREAK;
elsif OPMODE = SDLC then
    BREAK_ABORT_LATCH <= ABORT;
end if;

    DCD_LATCH <= not DCDn;
        end if;

-- 3.8.5 Clear to Send (taken from the AM8530 datasheet):
-- The CTS Status bit reports the state of the CTS input pin the 
-- last time any of the enabled External/Status bits changed. 
if CTS_D xor CTS = '1' then
    CTS <= not CTSn;
elsif DCD_IE = '1' and (DCD_D xor DCD = '1') then
    CTS <= not CTSn;
elsif EOM_IE = '1' and (Tx_UNDERRUN_EOM_D xor Tx_UNDERRUN_EOM = '1') then
    CTS <= not CTSn;
elsif SYNC_HUNT_IE = '1' and (SYNC_HUNT_D xor SYNC_HUNT = '1') then
    CTS <= not CTSn;
elsif ZCOUNT_IE = '1' and (ZCOUNT_D xor ZCOUNT = '1') then
    CTS <= not CTSn;
end if;

if CTS_IE = '1' and LATCH_LOCK = true and (CTS_D xor CTS = '1') then -- any edge detected.
    CTS_ODD := not CTS_ODD;

        -- This logic causes the respective signals in asynchronous or external sync mode to 
        -- release the interrupt on odd numbers of transistions. In detail this means, that if a
        -- rising edge is detected first during a pending service, the interrupt is released on 
        -- the following rising edge. The falling edge works in the same manner.
        if OPMODE = ASYNC or OPMODE =/ ASYNC and SYNC_CHAR = "11" then -- Asynchronous or external mode.
            if CTS = '1' and CTS_D = '0' and LATCH_LOCK = true then
                CTS_LH := true; -- Low to high transistion detected.
            elsif CTS = '0' and CTS_D = '1' and LATCH_LOCK = true then
                CTS_HL := true; -- Low to high transistion detected.
            elsif CTS = '0' and CTS_D = '0' and LATCH_LOCK = false then
                CTS_LH := false;
            elsif CTS = '1' and CTS_D = '1' and LATCH_LOCK = false then
                CTS_HL := false;        
            end if;

            if SYNC_HUNT = '1' and SYNC_HUNT_D = '0' and LATCH_LOCK = true then
                SYNC_HUNT_LH := true; -- Low to high transistion detected.
            elsif SYNC_HUNT = '0' and SYNC_HUNT_D = '1' and LATCH_LOCK = true then
                SYNC_HUNT_HL := true; -- Low to high transistion detected.
            elsif SYNC_HUNT = '0' and SYNC_HUNT_D = '0' and LATCH_LOCK = false then
                SYNC_HUNT_LH := false;
            elsif SYNC_HUNT = '1' and SYNC_HUNT_D = '1' and LATCH_LOCK = false then
                SYNC_HUNT_HL := false;        
            end if;
        else
            CTS_LH := false;
            CTS_HL := false;
            SYNC_HUNT_LH := false;
            SYNC_HUNT_HL := false;
        end if;


        if RESET = '1' or RES = '1' then
            SYNC_HUNT <= '0';
        elsif OPMODE = ASYNC and XTAL = '1' then -- Crystal oscillator enabled.
            SYNC_HUNT <= '0';
        elsif OPMODE = ASYNC then
            SYNC_HUNT <= SYNC_IN;
        elsif OPMODE /= ASYNC and SYNC_CHAR = "11" and XTAL = '1' else -- External sync mode, crystal oscillator enabled.
            SYNC_HUNT <= '0';
        elsif OPMODE /= ASYNC and SYNC_CHAR = "11" else -- External sync mode.
            SYNC_HUNT <= SYNC_IN;
        elsif Rx_EN = '0' then -- MSYNC, BSYNC, SDLC.
            SYNC_HUNT <= '1';
        elsif Rx_STATE /= Rx_HUNT and Rx_NSTATE = Rx_HUNT then -- MSYNC, BSYNC, SDLC.
            SYNC_HUNT <= '1';
        elsif Rx_STATE = Rx_HUNT and Rx_NSTATE /= Rx_HUNT then -- MSYNC, BSYNC, SDLC.
            SYNC_HUNT <= '0';
        end if;
-- DCD
-- CTS
-- ZERO_COUNT
-- BREAK_ABORT
-- Tx Underrun / EOM


if (OPMODE = MSYNC or OPMODE = BSYNC or OPMODE = SDLC) and SYNC_CHAR /= "01" then
    BREAK_ABORT <= '0';
elsif EXT_INT_EN = '1' and BREAK_ABORT_IE = '1' then
    BREAK_ABORT <= BREAK_ABORT_LATCH;
elsif OPMODE = SDLC then
    BREAK_ABORT <= ABORT;
elsif OPMODE = ASYNC then
    BREAK_ABORT <= BREAK;
end if;

if RESET = '1' or RES = '1' then -- Reset or channel reset.
    Tx_UNDERRUN_EOM <= '1';
elsif Tx_EN = '0' or TX_INH = '1' then -- Transmitter disabled.
    Tx_UNDERRUN_EOM <= '1';
elsif SEND_ABORT = '1' then -- Send abort command.
    Tx_UNDERRUN_EOM <= '1';
elsif RES_Tx_UR_EOM = '1' then -- Reset transmit underrun command.
    Tx_UNDERRUN_EOM <= '0';
elsif OPMODE = ASYNC then
    Tx_UNDERRUN_EOM <= '1';
elsif EXT_INT_EN = '1' and EOM_IE = '1' and (OPMODE = MSYNC or OPMODE = BSYNC or OPMODE = SDLC) then
    Tx_UNDERRUN_EOM <= EOM_LATCH;
elsif
    Tx_UNDERRUN_EOM <= '1';
end if;

if EXT_INT_EN = '1' and DCD_IE = '1' then
    DCD <= DCD_LATCH;
else
    DCD <= not DCDn;
end if;

    
        BREAK_ABORT_D := BREAK_ABORT;
        CTS_D := CTS;
        DCD_D := DCD;
        SYNC_HUNT_D := SYNC_HUNT;
        Tx_UNDERRUN_EOM_D := Tx_UNDERRUN_EOM;
        ZCOUNT_D := ZCOUNT;
    end process EXTERNAL_INTERRUPT_LATCHES;

    IO_CONTROL: process(CLK)
    variable W_REQ_P        : std_logic;
    variable DTR_REQ_P      : std_logic;
    variable W_REQ_N        : std_logic;
    variable DTR_REQ_N      : std_logic;
    variable DTR_REQ_DELAY  : std_logic_vector(2 downto 0);
    begin
        if CLK = '1' and CLK' event then
            if RESET = '1' or RES = '1' then
                W_REQ_P := '0';
            elsif DMA_REQ_MODE = "100" and WR8_WR = '1' and Tx_BUFFER_EMPTY = '0' then -- Wait on transmit mode.
                W_REQ_P := '1';
            elsif DMA_REQ_MODE = "101" and RR8_RD = '1' and Rx_BUFFER_EMPTY = '1' then -- Wait on receive mode.
                W_REQ_P := '1';
            elsif DMA_REQ_MODE = "110" and DTRn_REQ = '1' and Tx_BUFFER_EMPTY = '1' then -- DMA request on transmit mode.
                W_REQ_P := '1';
            elsif DMA_REQ_MODE = "111" and Rx_BUFFER_EMPTY = '0' then -- DMA request on receive mode.
                W_REQ_P := '1';
            else
                W_REQ_P := '0';
            end if;

            if RESET = '1' or RES = '1' then
                DTR_REQ_P := '0';
            elsif DMA_REQ_MODE = "110" and DTRn_REQ = '0' and Tx_BUFFER_EMPTY = '1' then -- DMA request on transmit mode using DMA_REQ.
                DTR_REQ_P := '1';
            end if;

            if DTR_REQ = '1' then
                DTR_REQ_DELAY := "111";
            else
                DTR_REQ_DELAY := DTR_REQ_DELAY(1 downto 0) & '0'; -- NMOS SCC delay, refer to the CMOS SCC data sheet.
            end if;
        end if;

        if CLK = '0' and CLK' event then
            W_REQ_N <= not W_REQ_P;
            DTR_REQ_N <= DTR_REQ_P;
        end if;

        if OPMODE = SDLC and SDLC_HDLC = '1' and FAST_DTR = '0' then
            Wn_REQn <= not W_REQ_N;
            DTRn_REQn <= not DTR_REQ_DELAY(2);
        elsif (OPMODE = MSYNC or OPMODE = BSYNC or OPMODE = SDLC) and DMA_REQ_MODE = "110" then
            Wn_REQn <= not W_REQ_N;
            DTRn_REQn <= not DTR_REQ_P;
        else
            Wn_REQn <= not W_REQ_N;
            DTRn_REQn <= not DTR_REQ_N;
        end if;
    end process IO_CONTROL;

-- Data section
BUFFER_IN
BUFFER_OUT <= BREAK_ABORT &  Tx_UNDERRUN_EOM & CTS & SYNC_HUNT & DCD & Tx_BUFFER_EMPTY & ZCOUNT & Rx_CHAR_AV when RR0_RD = '1' else
              Rx_ERROR when OPMODE = SCC and (FRAME_FIFO_EN = '0' or FRAME_FIFO_EMPTY = '1') and RR1_RD = '1' else -- Frame FIFO disabled or empty.
              '1' & FRAME_FIFO_DOUT(18 downto 17) & Rx_ERROR(4) & FRAME_FIFO_DOUT(16 downto 14) & Rx_ERROR(0) when OPMODE = SCC and RR1_RD = '1' else -- Frame FIFO enabled.
              FRAME_FIFO_DOUT(7 downto 0)) when RR6_RD = '1' else
              FRAME_FIFO_OVERFLOW & '1' & FRAME_FIFO_DOUT(13 downto 8) when RR7_RD = '1' else
              FIFO_REG(1)(15 downto 8) when RR8_RD = '1' else -- This is the character FIFO.
              FIFO_REG(1)(7 downto 0) when RR1_RD = '1' else x"00"; -- This is the Rx error FIFO.

-- Receiver section
    -- These are the receiver data input multiplexers.
    RxD_I <= TxD_I when LOOPBACK = '1' else RxD_D; -- This is the loopback multiplexer.

    RxD_SR <= RxD_I when OPMODE = ASYNC else
              RxD_NRZI; -- This is the NRZI multiplexer.

RxD_SDLC <= RxD_SR;

-- Hier auch Zero delete und Sync register.
    THREE_BIT_DELAY: process
    -- This is the trhee bit delay connected to the input
    -- of the receiver shift register.
    variable PIPE       : std_logic_vector(2 downto 0);
    begin
        wait until CLK = '1' and CLK' event;
        if Rx_STRB = '1' then
            PIPE <= RxD_SDLC & PIPE(1 downto 0);
            RxD_SR_3B <= PIPE(0);
        end if;
    end process THREE_BIT_DELAY;

    -- Error logic;
-- Rx_FIFO hält nicht ALL_SENT.
-- Hält Rx_FIFO die Residue bits?
    Rx_ERROR <= EOF & CRC_FRAME_ERR & Rx_OVR & PARITY_ERR & RESIDUE_BITS & ALL_SENT;
EOF

 -- No framing error for a BREAK condition is decoded by an empty shift register and the framing condition.
CRC_FRAME_ERR <= '1' when OPMODE = ASYNC and Rx_STATE = Rx_CHECK_FRAME and Rx_STRB = '1' and Rx_BITS = "11" RxD_SR_3B = '0' and Rx_SHIFTREG /= "000000000" else  -- 8 bits per character.
                 '1' when OPMODE = ASYNC and Rx_STATE = Rx_CHECK_FRAME and Rx_STRB = '1' and Rx_BITS = "01" RxD_SR_3B = '0' and Rx_SHIFTREG /= "000000000" else  -- 7 bits per character.
                 '1' when OPMODE = ASYNC and Rx_STATE = Rx_CHECK_FRAME and Rx_STRB = '1' and RxD_SR = '0' and Rx_SHIFTREG /= "000000000" else '0';
Rx_OVR
PARITY_ERR <= '1' when PAR_EN = '1' ad Rx_PARITY /= Rx_PARBIT else '0';
FIFO_PAR_ERR <= '1' when PAR_S_COND = '1' and FIFO_REG(1)(3) = '1' else '0'; -- Parity error.
RSIDUE_BITS...

    Rx_SCOND <= Rx_OVR or CRC_FRAME_ERR or EOF or FIFO_PAR_ERR;

    Rx_FIFO: process
-- This is the receiver FIFO. It is 16 bit wide...
    subtype T_01 is natural range 0 to 1;
    variable WRITE      : T_01;
    variable READ       : T_01;
    variable FIFO_LOCK  : boolean;
    begin
        wait until CLK = '1' and CLK' event;
        if RESET ='1' or RES = '1' then
            FIFO_WR_PNT <= 0;
        elsif CLK = '1' and CLK' event then
            if FIFO_WR = '1' and FIFO_LOCK = false then
                WRITE := 1;
            elsif FIFO_WR = '0' then
                WRITE := 0;
            end if;
            if FIFO_RD = '1' and FIFO_LOCK = false then
                READ := 1;
            elsif FIFO_RD = '0' then
                READ := 0;
            end if;
            if FIFO_WR_PNT = 3 and WRITE = 1 and READ = 0 then
                null; -- FIFO full, no further write.
            elsif FIFO_WR_PNT = 0 and WRITE = 0 and READ = 1 then
                null; -- FIFO empty, no further read.
            else
                FIFO_WR_PNT <= FIFO_WR_PNT + WRITE - READ;
            end if;
        end if;

        if RESET = '1' or RES = '1' then
            FIFO_REG <= (others => (others => '0'));
        else
            for i in 1 to 2 loop
                if i > FIFO_WR_PNT then
                    FIFO_REG(i) <= To_BitVector(Rx_SHIFTREG(7 downto 0) & Rx_ERROR);
                elsif FIFO_RD = '1' and FIFO_LOCK = false then
                    if i = ACSI_WR_PNT then
                        FIFO_REG(i) <= To_BitVector(Rx_SHIFTREG(7 downto 0) & Rx_ERROR);
                    end if;
                    FIFO_REG(i-1) <= FIFO_REG(i); -- Stage 0 is a pipeline stage.
                end if;
            end loop;
        end if;

        if RESET ='1' or RES = '1' then
            FIFO_LOCK := false;
        elsif RES_ERR = '1' then
            FIFO_LOCK := false;
        elsif Rx_INT_MODE /= "00" and FIFO_RD = '1' and FIFO_REG(2)(7 downto 0) /= x"00" then -- After FIFO_RD the error is on the top of the FIFO.
            FIFO_LOCK := true;
        end if;
    end process Rx_FIFO;

    FIFO_FULL <= '1' when FIFO_WR_PNT = 3 else '0';
    FIFO_EMPTY <= '1' when FIFO_WR_PNT = 0 else '0';

FIFO_RD -- nach RR8 Zugriff.
FIFO_WR <= '1' when OPMODE = ASYNC and Rx_STATE = Rx_CHECK_FRAME and Rx_STRB = '1' else '0';
FIFO_FULL
FIFO_EMPTY
    
-- Lesen und schreiben gleichzeitig?
-- Model the read and write as strobes.
    FRAME_FIFO_WR <= EOF;
    FRAME_FIFO_RD <= RR1_RD;

    FRAME_FIFO_CTRL: process
    variable ENTRY_CNT  : std_logic_vector(3 downto 0);
    begin
        wait until CLK = '1' and CLK' event;

        if RESET = '1' or RES = '1' then
            FRAME_FIFO_HEAD <= x"0";
            FRAME_FIFO_TAIL <= x"0";
            FRAME_FIFO_OVERFLOW <= '0';
            ENTRY_CNT := x"0";
        elsif OPMODE /= SDLC or FRAME_FIFO_EN = '0' then -- Frame FIFO disabled.
            FRAME_FIFO_HEAD <= x"0";
            FRAME_FIFO_TAIL <= x"0";
            FRAME_FIFO_OVERFLOW <= '0';
            ENTRY_CNT := x"0";
        else
-- PARITY wird auch eingelesen und verglichen...
            if FRAME_FIFO_WR = '1' and FRAME_FIFO_HEAD = x"9" then -- Rollover.
                FRAME_FIFO_HEAD <= x"0";
                ENTRY_CNT := ENTRY_CNT + '1';
            elsif FRAME_FIFO_WR = '1' then
                FRAME_FIFO_HEAD <= FRAME_FIFO_HEAD + '1'; -- One entry written.
                ENTRY_CNT := ENTRY_CNT + '1';
            end if;

            if FIFO_EMPTY = '1' then
                null; -- Do not read from an empty FIFO.
            elsif FRAME_FIFO_RD = '1' and FRAME_FIFO_TAIL = x"9" then -- Rollover.
                FRAME_FIFO_TAIL <= x"0";
                ENTRY_CNT := ENTRY_CNT - '1';
            elsif FRAME_FIFO_RD = '1' then
                FRAME_FIFO_TAIL <= FRAME_FIFO_TAIL + '1'; -- One entry written.
                ENTRY_CNT := ENTRY_CNT - '1';
            end if;

            if ENTRY_CNT > x"A" then
                FRAME_FIFO_OVERFLOW <= '1';
            end if;
        end if;

    end process FRAME_FIFO_CTRL;

    FRAME_FIFO_RAM: process(CLK, FIFO)
    variable ADR_PNTR  : integer range 0 to 15;
    begin
        if CLK = '1' and CLK' event then
            ADR_PNTR := To_Integer(unsigned(FRAME_FIFO_ADR));

            if OPMODE = SDLC and FRAME_FIFO_EN = '1' and FRAME_FIFO_WR = '1' then -- Frame FIFO enabled.
                FRAME_FIFO(ADR_PNTR) <= FRAME_FIFO_DIN;
            end if;
        end if;
        FRAME_FIFO_DOUT <= FRAME_FIFO(ADR_PNTR);
    end process FRAME_FIFO_RAM;

    FRAME_FIFO_DIN <= Rx_ERROR(6 downto 5) & Rx_ERROR(3 downto 1) & FRAME_BYTE_COUNTER
    FRAME_FIFO_ADR <= FRAME_FIFO_TAIL when FRAME_FIFO_RD = '1' else FRAME_FIFO_HEAD;

    FRAME_FIFO_EMPTY <= '1' when FRAME_FIFO_TAIL = FRAME_FIFO_HEAD else '0';

-- Verzögert, wenn FRAME_FIFO_WR = '1';
FRAME_BYTE_CLR <= ;

    P_FRAME_BYTE_COUNTER: process
    begin
        wait until CLK = '1' and CLK' event;
        if RESET = '1' or RES = '1' then
            FRAME_BYTE_COUNTER <= (others => '0');
        elsif OPMODE /= SDLC or FRAME_FIFO_EN = '0' then -- Frame FIFO disabled.
            FRAME_BYTE_COUNTER <= (others => '0');
        elsif FRAME_BYTE_CLR = '1' then
            FRAME_BYTE_COUNTER <= (others => '0');
        elsif FRAME_FIFO_WR = '1' then
            FRAME_BYTE_COUNTER <= FRAME_BYTE_COUNTER + '1';
        end if;
    end process P_FRAME_BYTE_COUNTER;

    NRZI_DECODER: process
    variable NRZI_IN        : std_logic;
    begin
        wait until CLK = '1' and CLK' event;
        if RESET = '1' or RES = '1' then
            NRZI_IN := '0';
        else
            if NRZ_FM = then -- NRZ.
                RxD_NRZI <= RxD_NRZI xnor NRZI_IN; -- NRZI decoding.
                NRZI_IN := RxD_I;
            elsif NRZ_FM = then -- NRZI.
RxD_NRZI <= RxD_NRZI xnor NRZI_IN; -- NRZI decoding.
NRZI_IN := RxD_I;
            elsif NRZ_FM = then -- FM1.
            else -- FM0.
            end if;
        end if;
    end process NRZI_DECODER;

	RECEIVER_SHIFTREG: process
    -- This is the receiver shift register. It is shifted in LSB first (right shift).
    variable BITCNT     : std_logic_vector(3 downto 0);
	begin
		wait until CLK = '1' and CLK' event;
		if Rx_STATE = IDLE then
			Rx_SHIFTREG <= (others => '0');
            BITCNT := x"0";
            Rx_SHFT_RDY <= '0';
            BREAK <= '0';
        elsif Rx_STATE = Rx_SHIFTIN = true and Rx_STRB = '1' then
-- Rx Bitfolge nochmal checken.
            BITCNT := BITCNT + '1';
            if PAR_EN = '1' then
                case Rx_BITS is
                    when "11" => -- 8 bits per character plus parity.
                        Rx_SHIFTREG <= RxD_SR_3B & Rx_SHIFTREG(8 downto 1);
                    when "01" => -- 7 bits per character plus parity.
                        Rx_SHIFTREG <= '0' & RxD_SR_3B & Rx_SHIFTREG(7 downto 1);
                    when "10" => -- 6 bits per character plus parity.
                        Rx_SHIFTREG <= "00" & RxD_SR & Rx_SHIFTREG(6 downto 1);
                    when others => -- 5 bits per character plus parity.
                        Rx_SHIFTREG <= "000" & RxD_SR & Rx_SHIFTREG(5 downto 1);
                end case;
            else
                case Rx_BITS is
                    when "11" => -- 8 bits per character.
                        Rx_SHIFTREG <= '0' & RxD_SR_3B & Rx_SHIFTREG(7 downto 1);
                    when "01" => -- 7 bits per character.
                        Rx_SHIFTREG <= "00" & RxD_SR_3B & Rx_SHIFTREG(6 downto 1);
                    when "10" => -- 6 bits per character.
                        Rx_SHIFTREG <= "000" & RxD_SR & Rx_SHIFTREG(5 downto 1);
                    when others => -- 5 bits per character.
                        Rx_SHIFTREG <= "0000" & RxD_SR & Rx_SHIFTREG(4 downto 1);
                end case;
            end if;
        end if;
        
        -- Be aware, that the 3 bit delay is taken into account in case of asynchronous
        -- 8 bit and 7 bit per character.
        if Rx_STRB = '1' and PAR_EN = '1' and Rx_BITS = "11" and BITCNT = x"C" then -- 8 bits per character plus parity.
            Rx_SHFT_RDY <= '1';
        elsif Rx_STRB = '1' and PAR_EN = '1' and Rx_BITS = "11" and BITCNT = x"B" then -- 7 bits per character plus parity.
            Rx_SHFT_RDY <= '1';
        elsif Rx_STRB = '1' and PAR_EN = '1' and Rx_BITS = "11" and BITCNT = x"7" then -- 6 bits per character plus parity.
            Rx_SHFT_RDY <= '1';
        elsif Rx_STRB = '1' and PAR_EN = '1' and Rx_BITS = "11" and BITCNT = x"6" then -- 5 bits per character plus parity.
            Rx_SHFT_RDY <= '1';
        elsif Rx_STRB = '1' and PAR_EN = '0' and Rx_BITS = "11" and BITCNT = x"B" then -- 8 bits per character without parity.
            Rx_SHFT_RDY <= '1';
        elsif Rx_STRB = '1' and PAR_EN = '0' and Rx_BITS = "11" and BITCNT = x"A" then -- 7 bits per character without parity.
            Rx_SHFT_RDY <= '1';
        elsif Rx_STRB = '1' and PAR_EN = '0' and Rx_BITS = "11" and BITCNT = x"6" then -- 6 bits per character without parity.
            Rx_SHFT_RDY <= '1';
        elsif Rx_STRB = '1' and PAR_EN = '0' and Rx_BITS = "11" and BITCNT = x"5" then -- 5 bits per character without parity.
            Rx_SHFT_RDY <= '1';
        end if;
        
        if Rx_STATE = Rx_CHECK_FRAME and OPMODE = ASYNC and Rx_SHIFTREG = "000000000" and Rx_STRB = '1' and Rx_BITS = "11" and RxD_SR_3B = '0' then -- 8 bits per character.
            BREAK <= '1'; -- Break condition detected.
        elsif Rx_STATE = Rx_CHECK_FRAME and OPMODE = ASYNC and Rx_SHIFTREG = "000000000" and Rx_STRB = '1' and Rx_BITS = "01" and RxD_SR_3B = '0' then -- 7 bits per character.
            BREAK <= '1'; -- Break condition detected.
        elsif Rx_STATE = Rx_CHECK_FRAME and OPMODE = ASYNC and Rx_SHIFTREG = "000000000" and Rx_STRB = '1' and Rx_BITS = "10" and RxD_SR = '0' then -- 6 bits per character.
            BREAK <= '1'; -- Break condition detected.
        elsif Rx_STATE = Rx_CHECK_FRAME and OPMODE = ASYNC and Rx_SHIFTREG = "000000000" and Rx_STRB = '1' and Rx_BITS = "00" and RxD_SR = '0' then -- 5 bits per character.
            BREAK <= '1'; -- Break condition detected.
        elsif OPMODE = ASYNC and Rx_STRB = '1' and Rx_BITS = "11" and RxD_SR_3B = '1' then -- 8 bits per character.
            BREAK <= '0'; -- Break condition detected.
        elsif OPMODE = ASYNC and Rx_STRB = '1' and Rx_BITS = "01" and RxD_SR_3B = '1' then -- 7 bits per character.
            BREAK <= '0'; -- Break condition detected.
        elsif OPMODE = ASYNC and Rx_STRB = '1' and Rx_BITS = "10" and RxD_SR = '1' then -- 6 bits per character.
            BREAK <= '0'; -- Break condition detected.
        elsif OPMODE = ASYNC and Rx_STRB = '1' and Rx_BITS = "00" and RxD_SR = '1' then -- 5 bits per character.
            BREAK <= '0'; -- Break condition detected.
        end if;
	end process RECEIVER_SHIFTREG;	

	Rx_PARITY_GEN: process(SHIFTREG, PAR_EN, PAR_EVEN_ODDn)
	variable EVEN	: std_logic;
    variable PAR01  : std_logic;
    variable PAR23  : std_logic;
    variable PAR45  : std_logic;
    variable PAR67  : std_logic;
    variable PAR03  : std_logic;
    variable PAR47  : std_logic;
	begin
        PAR01 := Rx_SHIFTREG(0) xor Rx_SHIFTREG(1);
        PAR23 := Rx_SHIFTREG(2) xor Rx_SHIFTREG(3);
        PAR45 := Rx_SHIFTREG(4) xor Rx_SHIFTREG(5);
        PAR67 := Rx_SHIFTREG(6) xor Rx_SHIFTREG(7);
        PAR03 := PAR01 xor PAR03;
        PAR47 := PAR45 xor PAR67;

        case Rx_BITS is
            when "11" => -- 8 bits per character plus parity.
                EVEN := PAR03 xor PAR47;
            when "01" => -- 7 bits per character plus parity.
                EVEN := PAR03 xor PAR45 xor SHIFTREG(6);
            when "10" => -- 6 bits per character plus parity.
                EVEN := PAR03 xor PAR45;
            when others => -- 5 bits per character plus parity.
                EVEN := PAR03 xor SHIFTREG(4);
        end case;

        if PAR_EN = '0' then
            Rx_PARITY <= '0';
        elsif PAR_EVEN_ODDn = '1' then -- Even parity.
            Rx_PARITY <= EVEN; 
        else -- Odd parity.
            Rx_PARITY <= not EVEN;
        end if;
	end process Rx_PARITY_GEN;

RECEIVER_CRC: process
begin
    wait until CLK = '1' and CLK' event;
    if RES_Rx_CRC = '1' and CRC_PRES = '1' then
        Rx_CRC_SHIFTREG <= (others => '1');
    elsif RES_Rx_CRC = '1' then
        Rx_CRC_SHIFTREG <= (others => '0');
    end if;
end process RECEIVER_CRC;

-- Transmitter section

    Tx_LEN <= 8 when Tx_BITS = "11" else
              7 when Tx_BITS = "01" else
              6 when Tx_BITS = "10" else
              5 when Tx_BITS = "00" and Tx_BUFFER(7 downto 5) = "000" else
              4 when Tx_BITS = "00" and Tx_BUFFER(7 downto 4) = "1000" else
              3 when Tx_BITS = "00" and Tx_BUFFER(7 downto 3) = "11000" else
              2 when Tx_BITS = "00" and Tx_BUFFER(7 downto 2) = "111000" else 
              1; -- Tx_BITS = "00" and Tx_BUFFER(7 downto 1) = "1111000".

    P_TRANSMIT_BUFFER: process
    variable NOT_EMPTY  : boolean;
    variable LOCK       : boolean;
    begin
        wait until CLK = '1' and CLK' event;
        if (RESET or RES) = '1' then
            Tx_BUFFER_EMPTY <= '1';
        elsif Tx_STATE = IDLE and Tx_NSTATE = Tx_SHIFTOUT then
            Tx_BUFFER_EMPTY <= '1';
elsif RES_TxINT = '1' then
        elsif WR8_WR = '1' and LOCK = false then
            Tx_BUFFER <= BUFFER_IN;
            SYNC_SDLC_ADR_I <= SYNC_SDLC_ADR;
            SYNC_SDLC_FLAG_I <= SYNC_SDLC_FLAG;
            Tx_BUFFER_EMPTY <= '0';
            LOCK := true;
        elsif WR8_WR = '0' then
            LOCK := false;
        end if;
    end process P_TRANSMIT_BUFFER;

    Tx_PARITY_GEN: process (PAR_EN, Tx_SR_DATAIN)
	variable EVEN	: std_logic;
    variable PAR01  : std_logic;
    variable PAR23  : std_logic;
    variable PAR45  : std_logic;
    variable PAR67  : std_logic;
    variable PAR03  : std_logic;
    variable PAR47  : std_logic;
	begin
        PAR01 := Tx_SR_DATAIN(0) xor Tx_SR_DATAIN(1);
        PAR23 := Tx_SR_DATAIN(2) xor Tx_SR_DATAIN(3);
        PAR45 := Tx_SR_DATAIN(4) xor Tx_SR_DATAIN(5);
        PAR67 := Tx_SR_DATAIN(6) xor Tx_SR_DATAIN(7);
        PAR03 := PAR01 xor PAR03;
        PAR47 := PAR45 xor PAR67;

        case Tx_LEN is
            when 8 => EVEN := PAR03 xor PAR47;
            when 7 => EVEN := PAR03 xor PAR45 xor Tx_SR_DATAIN(6);
            when 6 => EVEN := PAR03 xor PAR45;
            when 5 => EVEN := PAR03 xor Tx_SR_DATAIN(4);
            when 4 => EVEN := PAR03;
            when 3 => EVEN := PAR01 xor Tx_SR_DATAIN(2);
            when 2 => EVEN := PAR01;
            when 1 => EVEN := Tx_SR_DATAIN(0);
        end case;

        if PAR_EN = '0' then
            Tx_PARITY <= '0';
        elsif PAR_EVEN_ODDn = '1' then -- Even parity.
            Tx_PARITY <= EVEN; 
        else -- Odd parity.
            Tx_PARITY <= not EVEN;
        end if;
	end process Tx_PARITY_GEN;

    -- This is the transmit shift register parallel data. Be aware that the start bits '0' is adjusted right hand side and the
    -- stop bits '1' are adjusted left hand side.
    Tx_SR_DATAIN <= "0000000111" & Tx_PARITY & Tx_BUFFER & '0' when OPMODE = ASYNC and Tx_LEN = 8 and SYNC_MODE = "01" else
                    "00000000111" & Tx_PARITY & Tx_BUFFER(6 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 7 and PAR_EN = '1' else
                    "000000000111" & Tx_PARITY & Tx_BUFFER(5 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 6 and PAR_EN = '1' else
                    "0000000000111" & Tx_PARITY & Tx_BUFFER(4 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 5 and PAR_EN = '1' else
                    "00000000000111" & Tx_PARITY & Tx_BUFFER(3 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 4 and PAR_EN = '1' else
                    "000000000000111" & Tx_PARITY & Tx_BUFFER(2 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 3 and PAR_EN = '1' else
                    "0000000000000111" & Tx_PARITY & Tx_BUFFER(1 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 2 and PAR_EN = '1' else
                    "00000000000000111" & Tx_PARITY & Tx_BUFFER(0) & '0' when OPMODE = ASYNC and Tx_LEN = 1 and PAR_EN = '1'
                    "0000000111" & Tx_BUFFER & '0' when OPMODE = ASYNC and Tx_LEN = 8 and SYNC_MODE = "01" else
                    "00000000111" & Tx_BUFFER(6 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 7 else
                    "000000000111" & Tx_BUFFER(5 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 6 else
                    "0000000000111" & Tx_BUFFER(4 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 5 else
                    "00000000000111" & Tx_BUFFER(3 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 4 else
                    "000000000000111" & Tx_BUFFER(2 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 3 else
                    "0000000000000111" & Tx_BUFFER(1 downto 0) & '0' when OPMODE = ASYNC and Tx_LEN = 2 else
                    "00000000000000111" & Tx_BUFFER(0) & '0' when OPMODE = ASYNC and Tx_LEN = 1 else

	TRANSMITTER_SHIFTREG: process
    variable Tx_SHIFTREG    : std_logic_vector(19 downto 0);
    variable BITCNT         : integer range 0 to 15;
	begin
		wait until CLK = '1' and CLK' event;
		if Tx_STATE = IDLE then
			Tx_SHIFTREG := (others => '0');
            BITCNT := 0;
            Tx_SHFT_RDY <= '0';
        elsif Tx_STATE = IDLE and Tx_NSTATE = Tx_START then
            Tx_SHIFTREG := Tx_SR_DATAIN;
        elsif Tx_INH = '1' or SEND_BREAK = '1' then
            null;
        elsif Tx_STATE = Tx_START and  Tx_STRB = '1' then
            Tx_SHIFTREG := '0' Tx_SHIFTREG(19 downto 1); -- Shift right.
        elsif Tx_STATE = Tx_SHIFTOUT and  Tx_STRB = '1' then
            BITCNT := (BITCNT + 1) mod 16;
            Tx_SHIFTREG := '0' Tx_SHIFTREG(19 downto 1); -- Shift right.
        elsif Tx_STATE = Tx_STOP_1 and  Tx_STRB = '1' then
            Tx_SHIFTREG := '0' Tx_SHIFTREG(19 downto 1); -- Shift right.
        elsif Tx_STATE = Tx_STOP_2 and  Tx_STRB = '1' then
            Tx_SHIFTREG := '0' Tx_SHIFTREG(19 downto 1); -- Shift right.
        end if;
        
        if SEND_BREAK = '1' and Tx_STRB = '1' then
            TxD_SR <= '0';
        elsif OPMODE = ASYNC and Tx_STATE /= Tx_SHIFTOUT then
            TxD_SR <= '1';
        elsif Tx_STRB = '1' then
            TxD_SR <= Tx_SHIFTREG(0);
        end if;

        if OPMODE = ASYNC and PAR_EN = '1' and Tx_LEN = BITCNT then -- Character plus parity bit.
            Tx_SHFT_RDY <= '1';
        elsif OPMODE = ASYNC and PAR_EN = '0' and Tx_LEN + 1 = BITCNT then -- Character only.
            Tx_SHFT_RDY <= '1';
        end if;

        if (RESET or RES) = '1' then
            ALL_SENT <= '0';
        elsif OPMODE /= ASYNC then
            ALL_SENT <= '1';
        elsif WR8_WR = '1' then
            ALL_SENT <= '0';
        elsif OPMODE = ASYNC and Tx_STATE /= IDLE and Tx_NSTATE = IDLE and Tx_BUFFER_EMPTY = '1' then
            ALL_SENT <= '1';
        end if;
	end process TRANSMITTER_SHIFTREG;	

ZERO_INSERTION: process
--Tx_SR ist Eingang
--Tx_CRC ist Eingang
--Tx_ZI ist Ausgang
begin
    wait until CLK = '1' and CLK' event;
end process ZERO_INSERTION;

TRANSMITTER_CRC: process
--Tx_SR ist Eingang
--Tx_CRC ist Ausgang
begin
    wait until CLK = '1' and CLK' event;
    if RES_Tx_CRC = '1' and CRC_PRES = '1' then
        Tx_CRC_SHIFTREG <= (others => '1');
    elsif RES_Tx_CRC = '1' then
        Tx_CRC_SHIFTREG <= (others => '0');
    end if;
end process TRANSMITTER_CRC;

	TRANSMITTER_MUX: process
    -- The transmit multiplexer is two bit delayed respective
    -- to its data input.
	begin
		wait until CLK = '1' and CLK' event;
if OPMODE = SDLC and then Tx_MUX <= Tx_ZI;
elsif OPMODE = SDLC and then Tx_MUX <= Tx_CRC;
else Tx_MUX <= TxD_SR;
end if;
	end process TRANSMITTER_MUX;	

NRZI_ENCODER: process
--Tx_MUX ist Eingang
--Tx_NRZI ist Ausgang
begin
    wait until CLK = '1' and CLK' event;
end process NRZI_ENCODER;

    -- This is the final Transmitter multiplexer and tri state control.
    TxD <= TxD_I;
    TxD_I <= Tx_MUX when OPMODE = ASYNC else
TxD_NRZI when else
             RxD_D when AUTO_ECHO = '1' else
RxD; -- when .

    TxD_EN <= '0' when TxD_PULLED_HIGH = '1' and Tx_EN = '0' and OPMODE = SDLC and SDLC_HDLC = '1' and NRZ_FM = "01" else '1';
end architecture TRANSCEIVER;

#####################################################################################################################################################################################################
-- Rx_CLK bei fallender Flanke von RxD auf halbe Bitzelle justieren.
-- Receive fifo wird nicht durch reset gelöscht
-- Rx_OVR??

if (RESET or RES) = '1' then
    Rx_CHAR_AV <= '0';
    Tx_UNDERRUN_EOM <= '1';
    BREAK_ABORT <= '0';
    ALL_SENT <= '0';
    RES_CODE <= "011";
    PAR_ERR <= '0';
    Rx_OVR <= '0';
    EOF <= '0';
    LOOP_SEND <= '0';
    ON_LOOP <= '0';
end if;

--Tx_EMPTY <= '1' after last Stop bit has passed the TxD pin.
-- TR_FLAG_SENT <= '1' when last zero bit of closing flag leaves TxD

Rx_CHAR

if (RESET or RES) = '1' then
    Rx_CHAR_1 <= '0';
if gelesen then
    Rx_CHAR_1 <= '0';
elsif EN_INT_RxCHAR = '1' ausgelöst and Character then
    Rx_CHAR_1 <= '0';
end if;

-- ZERO delete
--DELAY 3 bits
--CRC Delay
--CRC Checker
-- Rx_SYNC in synchronous mode <= '1' when SYNC_CHARACTER found
-- BREAK, ABORT und SYNC_CHAR sind das gleiche.

-- 3.9.3.1 Tx_BUFFER_FULL Modellierung: siehe die letzten beiden Absätze.
-- 3.9.3.2 Tx_BUFFER_FULL Modellierung: siehe die letzten beiden Absätze.
-- 3.9.3.4 Rx_BUFFER_EMPTY Modellierung: siehe die letzten beiden Absätze.
